package be.cylab.java.wowa.training;


import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;

/**
 * Class Trainer.
 * Learn wowa weights based on training dataset
 */
public class Trainer {

    private final TrainerParameters parameters;

    /**
     * Trainer constructor.
     * @param parameters TrainerParameters
     *                   Parameters for trainer
     */
    public Trainer(final TrainerParameters parameters) {
        this.parameters = parameters;
    }

    /**
     * @param data
     * @param expected
     * @return
     */
    final public SolutionDistance run(final List<double[]> data,
                                final double[] expected
    ) {
        List<SolutionDistance> current_population =
                this.generateInitialPopulationAndComputeDistances(
                        this.getParameters().getPopulationSize(),
                        data,
                        expected
                );

        Collections.sort(current_population);
        SolutionDistance best_solution =
                this.findBestSolution(current_population);
        for (int generation = 0;
             generation < this.parameters.getMaxGenerationNumber();
             generation++
        ) {
            current_population = this.performReproduction(
                    current_population, data, expected);
            SolutionDistance best_solution_of_current_population =
                    this.findBestSolution(current_population);

            if (this.getParameters().getLogger() != null) {
                this.getParameters().getLogger().log(Level.INFO, "Test");
            }

            //We found a new best solution
            if (best_solution_of_current_population.getDistance()
                    < best_solution.getDistance()) {
                best_solution = best_solution_of_current_population;
            }

        }
        return best_solution;
    }

    /**
     * @param solutions
     * @return
     */
    private SolutionDistance findBestSolution(
            final List<SolutionDistance> solutions
    ) {
        SolutionDistance best_solution =
                new SolutionDistance(solutions.get(0).weights_p.length);

        for (SolutionDistance solution : solutions) {
            if (solution.getDistance() < best_solution.getDistance()) {
                best_solution = solution;
            }
        }
        return best_solution;
    }

    /**
     * @param number_of_weights
     * @param population_size
     * @return
     */
    private List<SolutionDistance> generateInitialPopulation(
            final int number_of_weights,
            final int population_size
    ) {
        List<SolutionDistance> population = new ArrayList<>();
        for (int i = 0; i < population_size; i++) {
            SolutionDistance solution = new SolutionDistance(number_of_weights);
            for (int j = 0; j < number_of_weights; j++) {
                solution.weights_p[j] = Math.random();
                solution.weights_w[j] = Math.random();
            }
            solution.normalize();
            population.add(solution);
        }
        return population;
    }

    /**
     * @param solutions
     * @param data
     * @param expected
     * @return
     */
    private List<SolutionDistance> computeDistances(
            final List<SolutionDistance> solutions,
            final List<double[]> data,
            final double[] expected
    ) {
        for (SolutionDistance solution : solutions) {
            solution.computeScoreTo(data, expected);
        }
        return solutions;

    }

    /**
     * @param solutions
     * @param selected_elements
     * @param count
     * @return
     */
    private List<SolutionDistance> rouletteWheelSelection(
            final List<SolutionDistance> solutions,
            final List<SolutionDistance> selected_elements,
            final int count
    ) {
        double max = Utils.findMaxDistance(solutions);
        double min = Utils.findMinDistance(solutions);
        double sum = Utils.sumTotalDistance(solutions);


        if (solutions.size() < count) {
            throw new IllegalArgumentException(
                    "Not enough elements in population to select "
                            + count + "parents"
            );
        }
        while (selected_elements.size() < count) {
            double tos = Math.random();
            double normalized_distance = 0;
            for (SolutionDistance solution : solutions) {
                normalized_distance = normalized_distance
                        + (max + min - solution.getDistance()) / sum;
                if (normalized_distance > tos) {
                    selected_elements.add(solution);
                    solutions.remove(solution);
                    break;
                }
            }
        }
        return selected_elements;
    }

    /**
     * @param solutions
     * @param selected_elements
     * @param count
     * @return
     */
    private List<SolutionDistance> tournamentSelection(
            final List<SolutionDistance> solutions,
            final List<SolutionDistance> selected_elements,
            final int count
    ) {

        while (selected_elements.size() < count) {
            Collections.sort(solutions);

            //Select random element in the population
            int solution1_index = Utils.randomInteger(0, solutions.size());
            SolutionDistance solution1 = solutions.get(solution1_index);

            //Select random element in the population
            int solution2_index = Utils.randomInteger(0, solutions.size());
            SolutionDistance solution2 = solutions.get(solution2_index);

            //Compare the two selected element and put the solution with
            // the smallest distance in selected list
            if (solution1.getDistance() < solution2.getDistance()) {
                selected_elements.add(solution1);
                solutions.remove(solution1);
            } else {
                selected_elements.add(solution2);
                solutions.remove(solution2);

            }
        }
        return selected_elements;
    }

    /**
     * @param solutions
     * @param count
     * @param selection_method
     * @return
     */
    private List<SolutionDistance> selectParents(
            final List<SolutionDistance> solutions,
            final int count,
            final int selection_method
    ) {
        List<SolutionDistance> selected_parents = new ArrayList<>();
        //Select the two best current solutions

        //Sort population
        Collections.sort(solutions);
        //Put the two best elements in selected parents list
        selected_parents.add(solutions.get(0));
        selected_parents.add(solutions.get(1));
        //Remove the two best elements from the list solutions
        solutions.remove(0);
        solutions.remove(1);

        if (selection_method == TrainerParameters.SELECTION_METHOD_RWS) {

            return this.rouletteWheelSelection(solutions,
                    selected_parents,
                    count - 2
            );
        } else if (selection_method == TrainerParameters.SELECTION_METHOD_TOS) {
            return this.tournamentSelection(solutions,
                    selected_parents,
                    count - 2
            );
        }
        throw new IllegalArgumentException("Invalid selection method");

    }

    /**
     * @param solutions
     * @return
     */
    private List<SolutionDistance> doReproduction(
            List<SolutionDistance> solutions
    ) {
        int nbr_weights = solutions.get(0).weights_p.length;

        //we add children to the current list of solutions
        while (solutions.size() < this.parameters.getPopulationSize()) {
            int dad = Utils.randomInteger(0, solutions.size() - 1);
            int mom;
            do {
                mom = Utils.randomInteger(0, solutions.size() - 1);
            } while (dad == mom);

            int cut_position = Utils.randomInteger(0, nbr_weights - 1);

            //beta is used to compute the new value at the cut position
            double beta = Math.random();
            this.reproduce(solutions.get(dad),
                    solutions.get(mom),
                    solutions,
                    cut_position,
                    beta
            );
        }
        while (solutions.size() > this.getParameters().getPopulationSize()) {
            solutions.remove(solutions.get(solutions.size() - 1));
        }
        return solutions;
    }

    /**
     * @param dad
     * @param mom
     * @param solutions
     * @param cutPosition
     * @param beta
     * @return
     */
    private void reproduce(final SolutionDistance dad,
                                             final SolutionDistance mom,
                                             List<SolutionDistance> solutions,
                                             int cutPosition,
                                             double beta
    ) {
        double p_new1W = dad.weights_w[cutPosition] - beta
                * (dad.weights_w[cutPosition] - mom.weights_w[cutPosition]);
        double p_new2W = mom.weights_w[cutPosition] + beta
                * (dad.weights_w[cutPosition] - mom.weights_w[cutPosition]);

        double p_new1P = dad.weights_p[cutPosition] - beta
                * (dad.weights_p[cutPosition] - mom.weights_p[cutPosition]);
        double p_new2P = mom.weights_p[cutPosition] + beta
                * (dad.weights_p[cutPosition] - mom.weights_p[cutPosition]);

        SolutionDistance child1 = new SolutionDistance(dad.weights_p.length);
        SolutionDistance child2 = new SolutionDistance(dad.weights_p.length);

        for (int i = 0; i < cutPosition; i++) {
            child1.weights_w[i] = dad.weights_w[i];
            child1.weights_p[i] = dad.weights_p[i];

            child2.weights_w[i] = mom.weights_w[i];
            child2.weights_p[i] = mom.weights_p[i];
        }

        child1.weights_w[cutPosition] = p_new1W;
        child2.weights_w[cutPosition] = p_new2W;
        child1.weights_p[cutPosition] = p_new1P;
        child2.weights_p[cutPosition] = p_new2P;

        int nbre_weights = dad.weights_w.length;
        for (int i = cutPosition + 1; i < nbre_weights; i++) {
            child1.weights_w[i] = mom.weights_w[i];
            child1.weights_p[i] = mom.weights_p[i];

            child2.weights_w[i] = dad.weights_w[i];
            child2.weights_p[i] = dad.weights_p[i];
        }
        //Check and correct only if we used a QuasiRandom generation
        child1.normalize();
        child2.normalize();
        solutions.add(child1);
        solutions.add(child2);

    }

    /**
     * @param solutions
     * @return
     */
    private List<SolutionDistance> randomlyMutateGenes(
            List<SolutionDistance> solutions
    ) {
        double probability = this.getParameters().getMutationRate() / 100;

        for (int i = 0; i < solutions.size(); i++) {
            if (i > 1) {
                solutions.get(i).randomlyMutateWithProbability(probability);
            }

        }
        return solutions;
    }

    /**
     * @return
     */
    private TrainerParameters getParameters() {
        return this.parameters;
    }

    /**
     * @param population_size
     * @param data
     * @param expected
     * @return
     */
    public List<SolutionDistance> generateInitialPopulationAndComputeDistances(
            int population_size,
            List<double[]> data,
            double[] expected
    ) {
        int numberOfWeights = data.get(0).length;
        List<SolutionDistance> initial_population =
                this.generateInitialPopulation(
                        numberOfWeights,
                        population_size
                );
        return this.computeDistances(initial_population, data, expected);
    }

    /**
     * @param population
     * @param data
     * @param expected
     * @return
     */
    private List<SolutionDistance> performReproduction(
            List<SolutionDistance> population,
            List<double[]> data,
            double[] expected
    ) {
        List<SolutionDistance> parents = this.selectParents(population,
                this.getParameters().getNumberOfParents(),
                this.getParameters().getSelectionMethod());
        List<SolutionDistance> new_generation = this.doReproduction(parents);
        List<SolutionDistance> mutated =
                this.randomlyMutateGenes(new_generation);

        return this.computeDistances(mutated, data, expected);
    }

    /**
     * @param child
     */
    private void checkAndCorrectNullWeightVector(SolutionDistance child) {

    }


}
